///////////////////////////////////////////////////////////
//                      MobSurgeon                       //
//           mob-            //
//           Copyright (C) 2005-2007 Gipat Group         //
//                             //
//    Gipat Group's opened EI-editor-utility license     //
//                       1.0                       //
//                                                       //
//                  www.gipatgroup.org                   //
///////////////////////////////////////////////////////////

//      , ....  :
// 1) Sagrer (sagrer@yandex.ru)

////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------


#pragma hdrstop

#include "MobFormat.h"
#include <Dialogs.hpp>
#include "initer.h"
#include "MainForm.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

TMobLoaderThread *MobLoaderThr;
TMobSaverThread *MobSaverThr;

TMob::TMob()
{
  //.
  UsingThread = false;
  Loading = false;
  Saving = false;
  SaveAttempts = 10;  //        10 .
  TailSize = 0;

  UnloadMob();  //   - .
}

TMob::~TMob()
{
  //.
  // 
  UnloadMob();

}

void TMob::UnloadMob()
//            .
{
  FullLogStr = "";
  LastLogMessage = "";
  NewMessageSent = false;
  UnknownBlocks.Length = 0;

  if (TailSize > 0)
  {
    // ,   .
    delete Tail;
    TailSize = 0;
  };
  MobLoaded = false;     //,       -    .
  CurrBlockSgn = false;

  //  ...
  ScriptBlockExists = false;   // ...
  ClearObjectsBlockInfo();     // ...

};

void TMob::ClearObjectsBlockInfo()
//       ...
{
  UnknObjs.Length = 0;
  UnitObjs.Length = 0;
  ObjectObjs.Length = 0;
  LeverObjs.Length = 0;
  TrapObjs.Length = 0;
  FlameObjs.Length = 0;
  Particle1Objs.Length = 0;
  Particle2Objs.Length = 0;
  Particle3Objs.Length = 0;

  ObjectsBlockExists = false;
};

bool TMob::LoadMob(String& FileName)
//   .
{
  bool Result = false;

  //    ...
  UnloadMob();

  Loading = true;  //     ...
  if (UsingThread == false)
  {
    //   
    Result = ReallyLoadMob(FileName);
  }
  else
  {
    //  .
    // ...
    MobLoaderThr = new TMobLoaderThread(true);
    //   ...
    MobLoaderThr->FileName = FileName;
    MobLoaderThr->PMobObj = this; //    TMob -   
                                  //   .
    //  .
    //LogEd->Lines->Clear();
    //  -      .
    // -      ...
    if (MainF->MainF_PK->ActivePage->Name == "LogSh")
    {
      LogEd->SetFocus();
    };  

    //  ...
    MobLoaderThr->Resume();

    //   ...
    do
    {
      Application->ProcessMessages();

      //,    ...
      if (NewMessageSent == true)
      {
        //    .  .
        LogEd->Lines->Add(LastLogMessage);
        NewMessageSent = false;
        //  ,    ....
        MobLoaderThr->Resume();
        
      };
    } while (MobLoaderThr->Finished != true);

    // ...
    Result = MobLoaderThr->result;

    UsingThread = false;
  };

  Loading = false;  //,   ...

  return Result;  //  ...

};

bool TMob::SaveMob(String& FileName)
//   .
{
  bool Result = false;

  //    ...
  FullLogStr = "";
  LastLogMessage = "";
  NewMessageSent = false;

  Saving = true;   //  ...
  if (UsingThread == false)
  {
    //   
    Result = ReallySaveMob(FileName);
  }
  else
  {
    //  .
    // ...
    MobSaverThr = new TMobSaverThread(true);
    //   ...
    MobSaverThr->FileName = FileName;
    MobSaverThr->PMobObj = this; //    TMob -   
                                  //   .
    //  .
    //  -      .
    if (MainF->MainF_PK->ActivePage->Name == "LogSh")
    {
      LogEd->SetFocus();
    };  

    //  ...
    MobSaverThr->Resume();

    //   ...
    do
    {
      Application->ProcessMessages();

      //,    ...
      if (NewMessageSent == true)
      {
        //    .  .
        LogEd->Lines->Add(LastLogMessage);
        NewMessageSent = false;
        //  ,    ....
        MobSaverThr->Resume();
        
      };
    } while (MobSaverThr->Finished != true);

    // ...
    Result = MobSaverThr->result;
    
    UsingThread = false;
  };
  Saving = false;  //, ...

  return Result;
};

void TMob::StringToLog(String& message)
//   .
{
  //    
  FullLogStr = FullLogStr+message+char(13)+char(10);

  //   -    :
  if (UsingThread == true)
  {
    if (Loading == true)
    {
      //  ...
      //   ,   
      MobLoaderThr->SynchrLogMess(message);
    };

    if (Saving == true)
    {
      //  ...
      //   ,   
      MobSaverThr->SynchrLogMess(message);
    };
  };
  
};

bool TMob::ReallyLoadMob(String& FileName)
//  .    .
{
  // ...
  DWORD ChetireBaita, BlockSignature;
  bool Result = false;
  bool AllOk = true;
  DWORD MainBlock_Size, CurrBlockSize = 0;
  __int64 Cursor = 0;
  __int64 FileSize, VosemBaitov = 0;
  int BlockNumber = 1;
  byte *TempBuff;

  //  -   -   :)
  StringToLog("");
  StringToLog("-----------------------------");
  //.     .
  TFileStream *File1;
  if (FileExists(FileName) == true)
  {
    //   .
    StringToLog(FileName);
    StringToLog(" ... ...");
    // ...
    TFileStream *File1;
    try
    {
      File1 = new TFileStream(FileName,fmOpenRead);
    }
    catch (...)
    {
      StringToLog("     !");
      AllOk = false;
      goto StopLoadMob_noloaded_Label;
    };

    //   .
    FileSize = File1->Size;
    byte *File1Loaded;
    File1Loaded = new byte[FileSize];

    //   
    File1->Read(File1Loaded,FileSize);
    // ...    .
    delete File1;

    //  ,  -       .

    //  - ,    .
    Cursor = 0;
    memcpy(&ChetireBaita,&File1Loaded[Cursor],4);
    Cursor = Cursor+4;
    if (ChetireBaita == MobSignature)
    {
      //   
      StringToLog("  ");

      //    -        :).

      //    .
      memcpy(&ChetireBaita,&File1Loaded[Cursor],4);
      Cursor = Cursor+4;
      MainBlock_Size = ChetireBaita;
      StringToLog("    "+IntToStr(MainBlock_Size)+" .");

      if (MainBlock_Size < FileSize)
      {
        //      -
        StringToLog("    ("+IntToStr(FileSize)+" ) -  ! :)");
        TailSize = FileSize-MainBlock_Size;  // -      :).
      };

      if (MainBlock_Size > FileSize)
      {
        //      -    -...
        //  ...
        StringToLog("    "+IntToStr(MainBlock_Size)+" . -    ! ("+IntToStr(FileSize)+" ).");
        AllOk = false;
        goto StopLoadMobLabel;
      };

      //        ...

      //  ...

      //  
      memcpy(&VosemBaitov,&File1Loaded[Cursor],8);
      Cursor = Cursor+8;
      if ((VosemBaitov != MainBlock_QuestMob) && (VosemBaitov != MainBlock_ZonalMob))
      {
        //-   .
        //   -     .
        StringToLog("    .  .");
        AllOk = false;
        goto StopLoadMobLabel;
      }
      else if (VosemBaitov == MainBlock_QuestMob)
      {
        // .
        MobType = "QuestMob";
        StringToLog("   .");
      }
      else if (VosemBaitov == MainBlock_ZonalMob)
      {
        // .
        MobType = "ZonalMob";
        StringToLog("   .");
      };

      //  ,  ...
      BlocksInMob = 1;

      //  -      :).      ,
      //  -   .
      do
      {
        //  .
        memcpy(&ChetireBaita,&File1Loaded[Cursor],4);
        Cursor = Cursor+4;
        BlockSignature = ChetireBaita;  //  
        BlocksInMob = BlocksInMob+1;    //  .

        //     ...
        if (BlockSignature == TextBlockSgn)
        {
          //  .
          BlockNumber = BlockNumber+1;  //    :).
          ScriptBlockExists = true;  //    :).

          //  
          memcpy(&ChetireBaita,&File1Loaded[Cursor],4);
          Cursor = Cursor+4;
          CurrBlockSize = ChetireBaita;  //   .

          //    :)
          TempBuff = new byte[CurrBlockSize];
          Cursor = Cursor-8;  //      ...
          memcpy(&TempBuff[0],&File1Loaded[Cursor],CurrBlockSize);
          Cursor = Cursor+CurrBlockSize;

          //    - .
          ReadScriptBlock(TempBuff,CurrBlockSize);

          // 
          delete TempBuff;

          //      ...
          StringToLog("   "+IntToStr(ScriptText.Length())+" .   "+IntToStr(BlockNumber)+".    "+IntToStr(CurrBlockSize)+" .");
        }

        else if (BlockSignature == ObjectsBlockSgn)
        {
          //  .
          BlockNumber = BlockNumber+1;  //    :).
          //       ,   :).

          //  
          memcpy(&ChetireBaita,&File1Loaded[Cursor],4);
          Cursor = Cursor+4;
          CurrBlockSize = ChetireBaita;  //   .

          //    :)
          TempBuff = new byte[CurrBlockSize];
          Cursor = Cursor-8;  //      ...
          memcpy(&TempBuff[0],&File1Loaded[Cursor],CurrBlockSize);
          Cursor = Cursor+CurrBlockSize;

          //    - .
          AllOk = ReadObjectsBlock(TempBuff,CurrBlockSize);
          if (AllOk == false)
          {
            //       -  .
            goto StopLoadMobLabel;
          };

          // 
          delete TempBuff;

          //      ...
          StringToLog("    "+IntToStr(CurrBlockSize)+" .   "+IntToStr(BlockNumber)+".");
          StringToLog("   "+IntToStr(UnitObjs.Length)+" .");
          StringToLog("   "+IntToStr(ObjectObjs.Length)+" .");
          StringToLog("   "+IntToStr(LeverObjs.Length)+" .");
          StringToLog("   "+IntToStr(TrapObjs.Length)+" .");
          StringToLog("   "+IntToStr(FlameObjs.Length)+" .");
          StringToLog("   "+IntToStr(Particle1Objs.Length)+"  1- .");
          StringToLog("   "+IntToStr(Particle2Objs.Length)+"  2- .");
          StringToLog("   "+IntToStr(Particle3Objs.Length)+"  3- .");
          StringToLog("   "+IntToStr(UnknObjs.Length)+"   .");
        }

        else
        {
          BlockNumber = BlockNumber+1;  //    :).
          //   .
          //     .

          UnknownBlocks.Length = UnknownBlocks.Length+1; //  .
          //  
          memcpy(&ChetireBaita,&File1Loaded[Cursor],4);
          Cursor = Cursor+4;

          //   ...
          UnknownBlocks[UnknownBlocks.High].BlockSize = ChetireBaita;
          UnknownBlocks[UnknownBlocks.High].BlockNumber = BlockNumber;
          UnknownBlocks[UnknownBlocks.High].BlockContent = new byte[ChetireBaita];
          Cursor = Cursor-8; //     .
          memcpy(&UnknownBlocks[UnknownBlocks.High].BlockContent[0],&File1Loaded[Cursor],ChetireBaita);
          Cursor = Cursor+ChetireBaita;

          //        :).
          StringToLog("   .    "+IntToStr(BlockNumber)+".");
        }; // else  

      } while (Cursor != MainBlock_Size);

      //  -     "" -       :).
      if (TailSize > 0)
      {
        Tail = new byte[TailSize];
        memcpy(&Tail[0],&File1Loaded[Cursor],TailSize);
        Cursor = Cursor+TailSize;
        StringToLog("   "+IntToStr(TailSize)+" .");
      };

      //  ...
      if (Cursor == FileSize)
      {
        //      -     :).
        StringToLog("     -       .");
        Result = true;
        MobLoaded = true;  //,     .
      };

    }
    else
    {
      //   !
      StringToLog("   !");
      AllOk = false;
    };

    StopLoadMobLabel:  //        
    ;
    // 
    //   ,   
    delete File1Loaded;
    StopLoadMob_noloaded_Label:   //  -     .
    ;
  };

  //  .  ,     
  // c\ .
  if (Result == true)
  {
    StringToLog("  !");
  }
  else
  {
    StringToLog("   !");
  };

  //     

  return Result;
};

bool TMob::ReallySaveMob(String& FileName)
//  .    .
{
  // ...
  DWORD ChetireBaita, BlockSignature;
  bool Result, BlockWrote, SgnFinded = false;
  bool AllOk = true, FileSaved = true;
  DWORD MainBlock_Size, CurrBlockSize, SizeAdress = 0;
  __int64 Cursor = 0;
  __int64 FileSize, VosemBaitov = 0;
  int BlockNumber = 1;
  int i;   // :).
  DynamicArray <byte> SaveBuffer;   //         ....
  byte *TempBuff;   //  ...
  DynamicArray <byte> TempBuff2;  //   ...

  //  ...

  //  ...
  CurrBlockSgn = 0;
  AllKnownBlocksSaved = false;

  //  -   -   :)
  StringToLog("");
  StringToLog("-----------------------------");
  //StringToLog(" ...");

  if (MobLoaded == true)
  {
    //          ...
    //  ...
    SaveBuffer.Length = Cursor+4;
    memcpy(&SaveBuffer[Cursor],&MobSignature,4);
    Cursor = Cursor+4;
    StringToLog(" ...");

    //     ...
    SaveBuffer.Length = Cursor+4;
    ChetireBaita = 0;
    memcpy(&SaveBuffer[Cursor],&ChetireBaita,4);
    Cursor = Cursor+4;

    //     -  . -...
    if (MobType == "QuestMob")
    {
      SaveBuffer.Length = Cursor+8;
      memcpy(&SaveBuffer[Cursor],&MainBlock_QuestMob,8);
      Cursor = Cursor+8;
      StringToLog("   ...");
    }
    else if (MobType == "ZonalMob")
    {
      SaveBuffer.Length = Cursor+8;
      memcpy(&SaveBuffer[Cursor],&MainBlock_ZonalMob,8);
      Cursor = Cursor+8;
      StringToLog("   ...");
    };

    //  -     .
    //    ,   ,
    //,          -
    //   :).

    do
    {
      //  .
      BlockNumber = BlockNumber+1;

      BlockWrote = false;   //....

      //,        ...
      for(i = 0;i<=UnknownBlocks.High;i++)
      {
        //     :).
        if (UnknownBlocks[i].BlockNumber == BlockNumber)
        {
          // ,       :).

          //  - ! :).
          SaveBuffer.Length = Cursor+UnknownBlocks[i].BlockSize;
          memcpy(&SaveBuffer[Cursor],&UnknownBlocks[i].BlockContent[0],UnknownBlocks[i].BlockSize);
          Cursor = Cursor+UnknownBlocks[i].BlockSize;
          StringToLog("  "+IntToStr(BlockNumber)+"    .");

          BlockWrote = true;  //    .

          i = UnknownBlocks.High;  //     ...
        };
      };

      // -       -   
      //,  " ".

      if (AllKnownBlocksSaved == false)
      {
        //      ...
        if (BlockWrote == false)
        {
          // -     ,   .

          // -    ...
          SgnFinded = false;
          do
          {
            //  ...
            CurrBlockSgn = GetNextBlockSgn(CurrBlockSgn);

            //,      .
            if (CurrBlockSgn == 0)
            {
              //   ...
              SgnFinded = true;
            }
            else if (CurrBlockSgn == TextBlockSgn)
            {
              // .
              if (ScriptBlockExists == true)
              {
                SgnFinded = true;
              };
            }
            else if (CurrBlockSgn == ObjectsBlockSgn)
            {
              // .
              if (ObjectsBlockExists == true)
              {
                SgnFinded = true;
              };
            };
          }
          while (SgnFinded != true);

          //  -    ,   "" ...
          if (CurrBlockSgn == 0)
          {
            //   ...
            //  ,       ...
            AllKnownBlocksSaved = true;
          }
          else if (CurrBlockSgn == TextBlockSgn)
          {
            // .

            // ...
            TempBuff2.Length = 0;

            //  -     :).
            GenScriptBlock(TempBuff2);

            //  -      :).
            SaveBuffer.Length = Cursor+TempBuff2.Length;
            memcpy(&SaveBuffer[Cursor],&TempBuff2[0],TempBuff2.Length);
            Cursor = Cursor+TempBuff2.Length;

            StringToLog("  "+IntToStr(BlockNumber)+"     "+IntToStr(TempBuff2.Length)+" .");
            BlockWrote = true;  //    .

            //   ...
            TempBuff2.Length = 0;

          }
          else if (CurrBlockSgn == ObjectsBlockSgn)
          {
            // .

            // ...
            TempBuff2.Length = 0;

            //  -     :).
            GenObjectsBlock(TempBuff2);

            //  -      :).
            SaveBuffer.Length = Cursor+TempBuff2.Length;
            memcpy(&SaveBuffer[Cursor],&TempBuff2[0],TempBuff2.Length);
            Cursor = Cursor+TempBuff2.Length;

            StringToLog("  "+IntToStr(BlockNumber)+"     "+IntToStr(TempBuff2.Length)+" .");
            BlockWrote = true;  //    .

            //   ...
            TempBuff2.Length = 0;
            
          };

        };  //if (BlockWrote == false)
      };  //if (AllKnownBlocksSaved == false)

    }
    while (BlockNumber != BlocksInMob);

    //      -     
    //(    -   :) ),     ...
    ChetireBaita = Cursor;
    memcpy(&SaveBuffer[4],&ChetireBaita,4);
    StringToLog("   ...");

    //.     .    
    // -    ,      -  
    // ... ...     :).

    if (TailSize > 0)
    {
      SaveBuffer.Length = Cursor+TailSize;
      memcpy(&SaveBuffer[Cursor],&Tail[0],TailSize);
      Cursor = Cursor+TailSize;
      StringToLog("    ,  "+IntToStr(TailSize)+" .");
    };

    // -     ...
    StringToLog("          "+IntToStr(Cursor)+" .");

    //  -     -     
    //   :).

    for (i = 0; i <= SaveAttempts; i++)
    {
      try
      {

        if (FileExists(FileName) == true)
        {
          //   -    .
          //       
          //   ...      -
          // -     -    :).

          if (DeleteFile(FileName) == true)
          {
            StringToLog(" "+FileName+" , ..      .");
          };
        };

        // -        -      :).

        //   ...
        TFileStream *File1;
        File1 = new TFileStream(FileName,fmCreate);

        //   ...
        File1->Write(&SaveBuffer[0],Cursor);

        // ...
        delete File1;

        if (FileExists(FileName) == true)
        {
          //      .
          StringToLog("      "+FileName+" .");
          Result = true; //  :).
          i = SaveAttempts;
        }
        else
        {
          //       .   :)
          // -    .
          StringToLog("  !!!!!        !!!!!!!!!!  !!!!!!");
          StringToLog("   ....");
          //        -   :).
        };
      }
      catch(...)
      {
        // - .

        //       .
        StringToLog("         "+FileName+" .");

        //        1 ...
        if (i < SaveAttempts)
        {
          StringToLog("   ....");
        };
      };
    };

    if ( (FileExists(FileName) == false) && (Result == false) )
    {
      //  10    .
      StringToLog("   !!!   !!!   !!!      -     !!!");
    };
  }
  else
  {
    //  ...   :).
    StringToLog("       !!!  !!!");
  };

  if (Result == true)
  {
    //  ...
    StringToLog("   :).");
  }
  else
  {
    //  ...
    StringToLog("    :(.");
  };

  return Result;
};

String TMob::DeCodeScript(byte *ScriptTextBytes,DWORD Size,DWORD Key)
//    .       .
{
  // ...
  String Result = "";
  DWORD TmpKey = 0;

  //   ...
  //  -    :)
  Result.SetLength(Size);
  memcpy(&Result[1],&ScriptTextBytes[0],Size);

  // -  .
  for (DWORD i=1; i<=Size; i++)
  {
    TmpKey = (((((Key*13) << 4 ) + Key) << 8 ) - Key);
    Key = Key + TmpKey*4 + 2531011;
    TmpKey = Key;
    TmpKey = TmpKey >> 16;
    Result[i] = Result[i] ^ TmpKey;
  };

  return Result;
};

byte* TMob::EnCodeScript(String ScriptTextStr,DWORD Size,DWORD Key)
//    .        .
{
  // ...
  byte *Result = new byte[Size];
  DWORD TmpKey = 0;

  //   ...
  //  -    :)
  memcpy(&Result[0],&ScriptTextStr[1],Size);

  // -  .
  for (DWORD i=0; i<Size; i++)
  {
    TmpKey = (((((Key*13) << 4 ) + Key) << 8 ) - Key);
    Key = Key + TmpKey*4 + 2531011;
    TmpKey = Key;
    TmpKey = TmpKey >> 16;
    Result[i] = Result[i] ^ TmpKey;
  };

  return Result;
};

DWORD TMob::GetNextBlockSgn(DWORD CurrSgn)
//      .
{
  //
  DWORD Result = 0;

  // ...
  if (CurrSgn == 0)
  {
    //     - .
    Result = TextBlockSgn;
  }
  else if (CurrSgn == TextBlockSgn)
  {
    //   -  -     1 .
    Result = ObjectsBlockSgn;
  }
  else if (CurrSgn == ObjectsBlockSgn)
  {
    //   -  -     1 .
    Result = 0;
  };

  return Result;
};

///////////////////////////////////
//   
///////////////////////////////////

bool TMob::ReadScriptBlock(byte *Buff, DWORD& Size)
// .
{
  // ...
  DWORD ChetireBaita;
  bool Result = false;
  __int64 Cursor = 0;
  byte *TempBuff;
  DWORD ScriptTextSize = Size - 12; //    :)

  ScriptText = "";  //   -  .

  Cursor = Cursor+8;

  // 
  memcpy(&ChetireBaita,&Buff[Cursor],4);
  Cursor = Cursor+4;
  ScriptKey = ChetireBaita;

  if (ScriptTextSize > 0)
  {
    //   -      :).
    TempBuff = new byte[ScriptTextSize];
    memcpy(&TempBuff[0],&Buff[Cursor],ScriptTextSize);
    Cursor = Cursor+ScriptTextSize;

    //      ...
    ScriptText = DeCodeScript(TempBuff,ScriptTextSize,ScriptKey);
    delete TempBuff;
  };

  Result = true;
  return Result;
};

bool TMob::ReadObjectsBlock(byte *Buff, DWORD& Size)
// .
{
  // ...
  DWORD ChetireBaita, CurrBlockSignature, ObjBlockSize, TempDW1,
  MinBlockSgn, MinBlockSize, TempDW2, i;
  bool Result = true;
  __int64 Cursor = 0, StartBlockPos = 0, UnknObjsNum = 0, UnitObjsNum = 0,
  ObjectObjsNum = 0, UnknObjsCurr = -1, UnitObjsCurr = -1, ObjectObjsCurr = -1,
  LeverObjsNum = 0, LeverObjsCurr = -1, TrapObjsNum = 0, TrapObjsCurr = -1,
  FlameObjsNum = 0, FlameObjsCurr = -1, Particle1ObjsNum = 0, Particle1ObjsCurr = -1,
  Particle2ObjsNum = 0, Particle2ObjsCurr = -1, Particle3ObjsNum = 0, Particle3ObjsCurr = -1;
  byte *TempBuff;
  byte TempByte;

  //,    ,   .  
  // ,    -    .
  //    -     .

  //.  .     ,    
  //  .

  Cursor = Cursor+8;  //     .

  //     ....    ...
  ClearObjectsBlockInfo();

  if (Size > 8)
  {
    //     .

    //   ...
    do
    {
      //   ...
      memcpy(&CurrBlockSignature,&Buff[Cursor],4);
      StartBlockPos = Cursor;  //    ...
      Cursor = Cursor+4;
      //  ...
      memcpy(&ObjBlockSize,&Buff[Cursor],4);
      Cursor = Cursor+4;
      //    :).
      Cursor = Cursor+ObjBlockSize-8;

      //   ...
      if (CurrBlockSignature == OB_Sgn_Unit)
      {
        //  ...
        UnitObjsNum = UnitObjsNum+1;
      }
      else if (CurrBlockSignature == OB_Sgn_Object)
      {
        //   ...
        ObjectObjsNum = ObjectObjsNum+1;
      }
      else if (CurrBlockSignature == OB_Sgn_Lever)
      {
        //  ...
        LeverObjsNum = LeverObjsNum+1;
      }
      else if (CurrBlockSignature == OB_Sgn_Trap)
      {
        //  ...
        TrapObjsNum = TrapObjsNum+1;
      }
      else if (CurrBlockSignature == OB_Sgn_Flame)
      {
        //  ...
        FlameObjsNum = FlameObjsNum+1;
      }
      else if (CurrBlockSignature == OB_Sgn_Particle1)
      {
        //   1- ...
        Particle1ObjsNum = Particle1ObjsNum+1;
      }
      else if (CurrBlockSignature == OB_Sgn_Particle2)
      {
        //   2- ...
        Particle2ObjsNum = Particle2ObjsNum+1;
      }
      else if (CurrBlockSignature == OB_Sgn_Particle3)
      {
        //   3- ...
        Particle3ObjsNum = Particle3ObjsNum+1;
      }
      else
      {
        //   .
        UnknObjsNum = UnknObjsNum+1;
      };
    }
    while (Cursor != Size);

    Cursor = 8;  //    -   :).
    //       .

    //   ...
    UnitObjs.Length = UnitObjsNum;
    ObjectObjs.Length = ObjectObjsNum;
    LeverObjs.Length = LeverObjsNum;
    TrapObjs.Length = TrapObjsNum;
    FlameObjs.Length = FlameObjsNum;
    Particle1Objs.Length = Particle1ObjsNum;
    Particle2Objs.Length = Particle2ObjsNum;
    Particle3Objs.Length = Particle3ObjsNum;
    UnknObjs.Length = UnknObjsNum;

    //     :).
    do
    {
      //   ...
      memcpy(&CurrBlockSignature,&Buff[Cursor],4);
      StartBlockPos = Cursor;  //    ...
      Cursor = Cursor+4;

      //   ...
      if (CurrBlockSignature == OB_Sgn_Unit)
      {
        //  ...
        //  ...
        UnitObjsCurr = UnitObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = UnitObjs[UnitObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = UnitObjs[UnitObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   "+IntToStr(UnitObjs[UnitObjsCurr].ObjectID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else if (CurrBlockSignature == OB_Sgn_Object)
      {
        //   ...
        //  ...
        ObjectObjsCurr = ObjectObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = ObjectObjs[ObjectObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = ObjectObjs[ObjectObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   "+IntToStr(ObjectObjs[ObjectObjsCurr].ObjectID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else if (CurrBlockSignature == OB_Sgn_Lever)
      {
        //  ...
        //  ...
        LeverObjsCurr = LeverObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = LeverObjs[LeverObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = LeverObjs[LeverObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   "+IntToStr(LeverObjs[LeverObjsCurr].ObjectID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else if (CurrBlockSignature == OB_Sgn_Trap)
      {
        //  ...
        //  ...
        TrapObjsCurr = TrapObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = TrapObjs[TrapObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = TrapObjs[TrapObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   "+IntToStr(TrapObjs[TrapObjsCurr].ObjectID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else if (CurrBlockSignature == OB_Sgn_Flame)
      {
        //  ...
        //  ...
        FlameObjsCurr = FlameObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = FlameObjs[FlameObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = FlameObjs[FlameObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   "+IntToStr(FlameObjs[FlameObjsCurr].ObjectID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else if (CurrBlockSignature == OB_Sgn_Particle1)
      {
        //   1- ...
        //  ...
        Particle1ObjsCurr = Particle1ObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = Particle1Objs[Particle1ObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = Particle1Objs[Particle1ObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   1-  "+IntToStr(Particle1Objs[Particle1ObjsCurr].ID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else if (CurrBlockSignature == OB_Sgn_Particle2)
      {
        //   2- ...
        //  ...
        Particle2ObjsCurr = Particle2ObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = Particle2Objs[Particle2ObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = Particle2Objs[Particle2ObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   2-  "+IntToStr(Particle2Objs[Particle2ObjsCurr].ID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else if (CurrBlockSignature == OB_Sgn_Particle3)
      {
        //   3- ...
        //  ...
        Particle3ObjsCurr = Particle3ObjsCurr+1;

        //  ...
        memcpy(&ObjBlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        TempBuff = new byte[ObjBlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&TempBuff[0],&Buff[Cursor],ObjBlockSize);
        Cursor = Cursor+ObjBlockSize;

        //     :)
        Result = Particle3Objs[Particle3ObjsCurr].ReadBlock(TempBuff,ObjBlockSize,true,this);
        //  ...
        if (Result == false)
        {
          goto StopLoadObjBlock;
        }
        else
        {
          // .  -     .
          TempDW1 = Particle3Objs[Particle3ObjsCurr].GetBlockSize();
          if (TempDW1 != ObjBlockSize)
          {
            //    .
            StringToLog("   3-  "+IntToStr(Particle3Objs[Particle3ObjsCurr].ID)+" - \"\" -  "+IntToStr(TempDW1)+"    "+IntToStr(ObjBlockSize)+" .");
          };
        };
      }

      else
      {
        //   .
        //  ...
        UnknObjsCurr = UnknObjsCurr+1;

        //  ...
        memcpy(&UnknObjs[UnknObjsCurr].BlockSize,&Buff[Cursor],4);
        Cursor = Cursor+4;

        //  .
        UnknObjs[UnknObjsCurr].BlockContent = new byte[UnknObjs[UnknObjsCurr].BlockSize];
        Cursor = Cursor-8;  //  .
        memcpy(&UnknObjs[UnknObjsCurr].BlockContent[0],&Buff[Cursor],UnknObjs[UnknObjsCurr].BlockSize);
        Cursor = Cursor+UnknObjs[UnknObjsCurr].BlockSize;

      };

    }
    while (Cursor != Size);
  };

  if (Result == true)
  {
    ObjectsBlockExists = true;   //  .
  };

  StopLoadObjBlock:        //     .
  ;
  return Result;
};

///////////////////////////////////
//   
///////////////////////////////////

bool TMob::GenScriptBlock(DynamicArray <byte>& Buff)
// .
{
  // ...
  DWORD ChetireBaita, BlockSignature;
  bool Result = false;
  DWORD Cursor = 0;
  byte *TempBuff;   //  ...

  // -    :).
  Buff.Length = 0;  // .

  //...
  Buff.Length = Cursor+4;
  memcpy(&Buff[Cursor],&TextBlockSgn,4);
  Cursor = Cursor+4;

  //  -    -  .
  Buff.Length = Cursor+4;
  ChetireBaita = 0;
  memcpy(&Buff[Cursor],&ChetireBaita,4);
  Cursor = Cursor+4;

  //...
  Buff.Length = Cursor+4;
  if (ScriptKey != 0)
  {
    ScriptKey = 32167;
  };
  memcpy(&Buff[Cursor],&ScriptKey,4);
  Cursor = Cursor+4;

  // ...
  //       -  :)
  if (ScriptText.Length() > 0)
  {
    //...
    if (ScriptKey != 0)
    {
      //   
      TempBuff = EnCodeScript(ScriptText,ScriptText.Length(),ScriptKey);
      // ...
      Buff.Length = Cursor+ScriptText.Length();
      memcpy(&Buff[Cursor],&TempBuff[0],ScriptText.Length());
      Cursor = Cursor+ScriptText.Length();

      delete TempBuff;  //  ,    .
                        //      EnCodeScript
    }
    else
    {
      // .    .
      Buff.Length = Cursor+ScriptText.Length();
      memcpy(&Buff[Cursor],&ScriptText[1],ScriptText.Length());
      Cursor = Cursor+ScriptText.Length();
    };
  };  

  //  ...
  memcpy(&Buff[4],&Cursor,4);

  // ...
  Result = true;

  return Result;
};

bool TMob::GenObjectsBlock(DynamicArray <byte>& Buff)
// .
{
  // ...
  int i = 0; //
  DWORD ChetireBaita, BlockSignature;
  bool Result = false;
  DWORD Cursor = 0;
  byte *TempBuff;   //  ...

  // -   :).
  Buff.Length = 0;  // .

  //...
  Buff.Length = Cursor+4;
  memcpy(&Buff[Cursor],&ObjectsBlockSgn,4);
  Cursor = Cursor+4;

  //  -    -  .
  Buff.Length = Cursor+4;
  ChetireBaita = 0;
  memcpy(&Buff[Cursor],&ChetireBaita,4);
  Cursor = Cursor+4;

  //  -      .

  // ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<UnitObjs.Length; i++)
  {
    ChetireBaita = ChetireBaita + UnitObjs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<UnitObjs.Length; i++)
  {
    TempBuff = UnitObjs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  // ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<ObjectObjs.Length; i++)
  {
    ChetireBaita = ChetireBaita + ObjectObjs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<ObjectObjs.Length; i++)
  {
    TempBuff = ObjectObjs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  // ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<LeverObjs.Length; i++)
  {
    ChetireBaita = ChetireBaita + LeverObjs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<LeverObjs.Length; i++)
  {
    TempBuff = LeverObjs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  // ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<TrapObjs.Length; i++)
  {
    ChetireBaita = ChetireBaita + TrapObjs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<TrapObjs.Length; i++)
  {
    TempBuff = TrapObjs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  // ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<FlameObjs.Length; i++)
  {
    ChetireBaita = ChetireBaita + FlameObjs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<FlameObjs.Length; i++)
  {
    TempBuff = FlameObjs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  //  1- ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<Particle1Objs.Length; i++)
  {
    ChetireBaita = ChetireBaita + Particle1Objs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<Particle1Objs.Length; i++)
  {
    TempBuff = Particle1Objs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  //  2- ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<Particle2Objs.Length; i++)
  {
    ChetireBaita = ChetireBaita + Particle2Objs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<Particle2Objs.Length; i++)
  {
    TempBuff = Particle2Objs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  //  3- ...
  // ...
  ChetireBaita = 0;
  for (i=0; i<Particle3Objs.Length; i++)
  {
    ChetireBaita = ChetireBaita + Particle3Objs[i].GetBlockSize();
  };
  Buff.Length = Cursor+ChetireBaita;

  //  -   ...
  for (i=0; i<Particle3Objs.Length; i++)
  {
    TempBuff = Particle3Objs[i].GenBlock(ChetireBaita);
    memcpy(&Buff[Cursor],&TempBuff[0],ChetireBaita);
    Cursor = Cursor+ChetireBaita;
    delete TempBuff;  // .
  };

  //   .
  //  (   )    
  //  ...
  ChetireBaita = 0;
  for (i=0; i<UnknObjs.Length; i++)
  {
    ChetireBaita = ChetireBaita+UnknObjs[i].BlockSize;
  };

  //   .       
  // -    :).
  Buff.Length = Cursor+ChetireBaita;

  for (i=0; i<UnknObjs.Length; i++)
  {
    //     .
    memcpy(&Buff[Cursor],&UnknObjs[i].BlockContent[0],UnknObjs[i].BlockSize);
    Cursor = Cursor+UnknObjs[i].BlockSize;
  };

  //  ...
  memcpy(&Buff[4],&Cursor,4);

  // ...
  Result = true;

  return Result;
};

////////////////////////////////////
//                //
////////////////////////////////////

void TMob::KillUnit (const int ObjNumber)
// .
{
  //
  int i;

  if (UnitObjs.Length > 0)
  {
    //    -  .
    if (ObjNumber < UnitObjs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < UnitObjs.Length-1; i++)
      {
        UnitObjs[i] = UnitObjs[i+1];
      };
    };

    //   1.
    UnitObjs.Length = UnitObjs.Length-1;
  };
};

void TMob::KillObject (const int ObjNumber)
// .
{
  //
  int i;

  if (ObjectObjs.Length > 0)
  {
    //    -  .
    if (ObjNumber < ObjectObjs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < ObjectObjs.Length-1; i++)
      {
        ObjectObjs[i] = ObjectObjs[i+1];
      };
    };

    //   1.
    ObjectObjs.Length = ObjectObjs.Length-1;
  };
};

void TMob::KillLever (const int ObjNumber)
// .
{
  //
  int i;

  if (LeverObjs.Length > 0)
  {
    //    -  .
    if (ObjNumber < LeverObjs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < LeverObjs.Length-1; i++)
      {
        LeverObjs[i] = LeverObjs[i+1];
      };
    };

    //   1.
    LeverObjs.Length = LeverObjs.Length-1;
  };
};

void TMob::KillTrap (const int ObjNumber)
// .
{
  //
  int i;

  if (TrapObjs.Length > 0)
  {
    //    -  .
    if (ObjNumber < TrapObjs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < TrapObjs.Length-1; i++)
      {
        TrapObjs[i] = TrapObjs[i+1];
      };
    };

    //   1.
    TrapObjs.Length = TrapObjs.Length-1;
  };
};

void TMob::KillFlame (const int ObjNumber)
// .
{
  //
  int i;

  if (FlameObjs.Length > 0)
  {
    //    -  .
    if (ObjNumber < FlameObjs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < FlameObjs.Length-1; i++)
      {
        FlameObjs[i] = FlameObjs[i+1];
      };
    };

    //   1.
    FlameObjs.Length = FlameObjs.Length-1;
  };
};

void TMob::KillParticle1 (const int ObjNumber)
//  1- .
{
  //
  int i;

  if (Particle1Objs.Length > 0)
  {
    //    -  .
    if (ObjNumber < Particle1Objs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < Particle1Objs.Length-1; i++)
      {
        Particle1Objs[i] = Particle1Objs[i+1];
      };
    };

    //   1.
    Particle1Objs.Length = Particle1Objs.Length-1;
  };
};

void TMob::KillParticle2 (const int ObjNumber)
//  2- .
{
  //
  int i;

  if (Particle2Objs.Length > 0)
  {
    //    -  .
    if (ObjNumber < Particle2Objs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < Particle2Objs.Length-1; i++)
      {
        Particle2Objs[i] = Particle2Objs[i+1];
      };
    };

    //   1.
    Particle2Objs.Length = Particle2Objs.Length-1;
  };
};

void TMob::KillParticle3 (const int ObjNumber)
//  3- .
{
  //
  int i;

  if (Particle3Objs.Length > 0)
  {
    //    -  .
    if (ObjNumber < Particle3Objs.Length-1)
    {
      //          .

      for (i = ObjNumber; i < Particle3Objs.Length-1; i++)
      {
        Particle3Objs[i] = Particle3Objs[i+1];
      };
    };

    //   1.
    Particle3Objs.Length = Particle3Objs.Length-1;
  };
};

//////////////////////////////////////////////////
//        .    //
//////////////////////////////////////////////////

bool TMob::CheckIDExists (const DWORD ID, bool DeleteIfExists)
//     ID   .     -  .
{
  //
  bool Result = true;
  int i;

  //.

  //.
  if (UnitObjs.Length > 0)
  {
    for (i = 0; ( (i < UnitObjs.Length) && (Result == true) ); i++)
    {
      if (UnitObjs[i].ObjectID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillUnit(i);
          if (i < UnitObjs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  //.
  if (ObjectObjs.Length > 0)
  {
    for (i = 0; ( (i < ObjectObjs.Length) && (Result == true) ); i++)
    {
      if (ObjectObjs[i].ObjectID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillObject(i);
          if (i < ObjectObjs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  //.
  if (LeverObjs.Length > 0)
  {
    for (i = 0; ( (i < LeverObjs.Length) && (Result == true) ); i++)
    {
      if (LeverObjs[i].ObjectID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillLever(i);
          if (i < LeverObjs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  //.
  if (TrapObjs.Length > 0)
  {
    for (i = 0; ( (i < TrapObjs.Length) && (Result == true) ); i++)
    {
      if (TrapObjs[i].ObjectID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillTrap(i);
          if (i < TrapObjs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  //.
  if (FlameObjs.Length > 0)
  {
    for (i = 0; ( (i < FlameObjs.Length) && (Result == true) ); i++)
    {
      if (FlameObjs[i].ObjectID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillFlame(i);
          if (i < FlameObjs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  // 1- .
  if (Particle1Objs.Length > 0)
  {
    for (i = 0; ( (i < Particle1Objs.Length) && (Result == true) ); i++)
    {
      if (Particle1Objs[i].ID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillParticle1(i);
          if (i < Particle1Objs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  // 2- .
  if (Particle2Objs.Length > 0)
  {
    for (i = 0; ( (i < Particle2Objs.Length) && (Result == true) ); i++)
    {
      if (Particle2Objs[i].ID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillParticle2(i);
          if (i < Particle2Objs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  // 3- .
  if (Particle3Objs.Length > 0)
  {
    for (i = 0; ( (i < Particle3Objs.Length) && (Result == true) ); i++)
    {
      if (Particle3Objs[i].ID == ID)
      {
        //  .
        if (DeleteIfExists == true)
        {
          KillParticle3(i);
          if (i < Particle3Objs.Length)
          {
            i--;   // i            .
          };
        }
        else
        {
          Result = false;  //    -   .
        };
      };  
    };
  };

  // .
  return Result;
};

DWORD TMob::FindCleanID(const bool IdForNPC)
//  ID.
{
  //
  DWORD Result, MaxID;
  bool Finded;
  int i;

  if (IdForNPC == true)
  {
    //     NPC-.
    Result = 1000000000;
    MaxID = 4294967295;
  }
  else
  {
    Result = 1;
    MaxID = 999999999;
  };

  Finded = true;
  //  :).
  for (Result; ( (Result <= MaxID) && (Finded == true) ); Result++)
  {
    Finded = false;  //       id - 
                     //      .

    //.
    if (UnitObjs.Length > 0)
    {
      for (i = 0; ( (i < UnitObjs.Length) && (Finded == false) ); i++)
      {
        if (UnitObjs[i].ObjectID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

    //.
    if (ObjectObjs.Length > 0)
    {
      for (i = 0; ( (i < ObjectObjs.Length) && (Finded == false) ); i++)
      {
        if (ObjectObjs[i].ObjectID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

    //.
    if (LeverObjs.Length > 0)
    {
      for (i = 0; ( (i < LeverObjs.Length) && (Finded == false) ); i++)
      {
        if (LeverObjs[i].ObjectID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

    //.
    if (TrapObjs.Length > 0)
    {
      for (i = 0; ( (i < TrapObjs.Length) && (Finded == false) ); i++)
      {
        if (TrapObjs[i].ObjectID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

    //.
    if (FlameObjs.Length > 0)
    {
      for (i = 0; ( (i < FlameObjs.Length) && (Finded == false) ); i++)
      {
        if (FlameObjs[i].ObjectID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

    // 1- .
    if (Particle1Objs.Length > 0)
    {
      for (i = 0; ( (i < Particle1Objs.Length) && (Finded == false) ); i++)
      {
        if (Particle1Objs[i].ID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

    // 2- .
    if (Particle2Objs.Length > 0)
    {
      for (i = 0; ( (i < Particle2Objs.Length) && (Finded == false) ); i++)
      {
        if (Particle2Objs[i].ID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

    // 3- .
    if (Particle3Objs.Length > 0)
    {
      for (i = 0; ( (i < Particle3Objs.Length) && (Finded == false) ); i++)
      {
        if (Particle3Objs[i].ID == Result)
        {
          //  .
          Finded = true;
        };  
      };
    };

  };

  //         -  
  //  .
  Result--;

  if ( ((Result == 999999999) || (Result == 4294967295)) && (Finded == true) )
  {
    //  ID   
    Result = 0;
  };

  // 
  return Result;
};


///////////////////////////////////
//TMobLoaderThread
///////////////////////////////////

//   ""  
// -     ...

__fastcall TMobLoaderThread::TMobLoaderThread(bool CreateSuspended)
  : TThread(CreateSuspended)
{
  //  .
};
//---------------------------------------------------------------------------
void __fastcall TMobLoaderThread::Execute()
{
  // ...
  Finished = false;
  result = PMobObj->ReallyLoadMob(FileName);
  this->Synchronize(SynchrFinished);
};

void TMobLoaderThread::SynchrLogMess(String& message)
//     .
{
  this->message = message;
  this->Synchronize(SynchrMessFunct);
  this->Suspend();  //    ,  
                    //  .   ...
};

void __fastcall TMobLoaderThread::SynchrMessFunct(void)
//  .
{
  PMobObj->LastLogMessage = message;
  PMobObj->NewMessageSent = true;   //.    .
};

void __fastcall TMobLoaderThread::SynchrFinished(void)
//,     :).
{
  Finished = true;
};

///////////////////////////////////
//TMobSaverThread
///////////////////////////////////

//   ""  
// -     ...

__fastcall TMobSaverThread::TMobSaverThread(bool CreateSuspended)
  : TThread(CreateSuspended)
{
  //  .
};
//---------------------------------------------------------------------------
void __fastcall TMobSaverThread::Execute()
{
  // ...
  Finished = false;
  result = PMobObj->ReallySaveMob(FileName);
  this->Synchronize(SynchrFinished);
};

void TMobSaverThread::SynchrLogMess(String& message)
//     .
{
  this->message = message;
  this->Synchronize(SynchrMessFunct);
  this->Suspend();  //    ,  
                    //  .   ...
};

void __fastcall TMobSaverThread::SynchrMessFunct(void)
//  .
{
  PMobObj->LastLogMessage = message;
  PMobObj->NewMessageSent = true;   //.    .
};

void __fastcall TMobSaverThread::SynchrFinished(void)
//,     :).
{
  Finished = true;
};

//////////////////////////////////////
//                     //
//////////////////////////////////////

DWORD GetNMobBlocksSize(byte *Buff, DWORD StartPosition, DWORD NBlocks)
//    N    ,     -  1  .
{
  // 
  DWORD BlockNum, Result = 0, BlockSize, Cursor = StartPosition;

  //...
  for (BlockNum = 1; BlockNum <= NBlocks; BlockNum++)
  {
    Cursor = Cursor+4;  //     ...
    memcpy(&BlockSize,&Buff[Cursor],4);  //  ...
    Cursor = Cursor+BlockSize-4;  //    ...
  };

  // ...
  Result = Cursor-StartPosition;

  return Result;
};

