///////////////////////////////////////////////////////////
//                      MobSurgeon                       //
//           mob-            //
//           Copyright (C) 2005-2007 Gipat Group         //
//                             //
//    Gipat Group's opened EI-editor-utility license     //
//                       1.0                       //
//                                                       //
//                  www.gipatgroup.org                   //
///////////////////////////////////////////////////////////

//      , ....  :
// 1) Sagrer (sagrer@yandex.ru)

////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------
//     *.mob-
//   .
#ifndef MobFormatH
#define MobFormatH
//---------------------------------------------------------------------------
//#include "StructMob.h"
#include <ComCtrls.hpp>

//      ...
#include "MobUnitObj.h"
#include "MobObjectObj.h"
#include "MobLeverObj.h"
#include "MobTrapObj.h"
#include "MobFlameObj.h"
#include "MobParticle1Obj.h"
#include "MobParticle2Obj.h"
#include "MobParticle3Obj.h"

// 

//  
const DWORD MobSignature = 0xA000; //(40960 "00 A0 00 00")   
const unsigned __int64 MainBlock_QuestMob = 34359791616; //(00 D0 00 00 08 00 00 00)    
const unsigned __int64 MainBlock_ZonalMob = 34359787520; //(00 0 00 00 08 00 00 00)    
const DWORD TextBlockSgn = 2899242187;  //(CB EC CE AC) -   .
const DWORD ObjectsBlockSgn = 45056;  //(00 B0 00 00) -   .

//   .
const DWORD OB_Sgn_Unit = 3149594624;     //(00 00 BB BB)  //.
const DWORD OB_Sgn_Object = 45057;     //(01 B0 00 00)  //.
const DWORD OB_Sgn_Lever = 3148611584;    //(00 00 AC BB)  //.
const DWORD OB_Sgn_Trap = 3148546048;     //(00 00 AB BB)  //.
const DWORD OB_Sgn_Flame = 3149856768;     //(00 00 BF BB)  //"".
const DWORD OB_Sgn_Particle1 = 43521;     //(01 AA 00 00)  //  1.
const DWORD OB_Sgn_Particle2 = 52225;     //(01 CC 00 00)  //  2.
const DWORD OB_Sgn_Particle3 = 56577;     //(01 DD 00 00)  //  3.

struct SUnknBlock       //  .
{
  __int64 BlockSize;    // .
  int BlockNumber;
  byte *BlockContent;   // .
};

struct SUnknObjBlock   //     .
{
  DWORD BlockSize;     // 
  byte *BlockContent;  // .
};

//  ...
class TUnitObjBlock;
//class TObjectObjBlock;

//  ...
class TMob
{
  public:
    //  ...
    String FullLogStr, LastLogMessage;
    bool UsingThread, NewMessageSent, Loading, Saving, MobLoaded,
    AllKnownBlocksSaved;
    TRichEdit *LogEd;    //     .
    int BlocksInMob, SaveAttempts;
    DWORD CurrBlockSgn;  //    ,   

    //   .
    String MobType;      // 

    // 
    DynamicArray <SUnknBlock> UnknownBlocks;   //    .

    //
    bool ScriptBlockExists;   //  
    DWORD ScriptKey;     //  
    String ScriptText;   //  ...

    //.
    bool ObjectsBlockExists;  //  
    DynamicArray <SUnknObjBlock> UnknObjs;  //      .
    DynamicArray <TUnitObjBlock> UnitObjs;   //      .
    DynamicArray <TObjectObjBlock> ObjectObjs;   //      .
    DynamicArray <TLeverObjBlock> LeverObjs;   //      .
    DynamicArray <TTrapObjBlock> TrapObjs;   //      .
    DynamicArray <TFlameObjBlock> FlameObjs;   //      .
    DynamicArray <TParticle1ObjBlock> Particle1Objs;   //        1.
    DynamicArray <TParticle2ObjBlock> Particle2Objs;   //        2.
    DynamicArray <TParticle3ObjBlock> Particle3Objs;   //        3.

    byte *Tail;          //  .
    __int64 TailSize;    //    .

    // 
    TMob();
    ~TMob();
    bool LoadMob(String& FileName);  //   .  ReallyLoadMob
    bool SaveMob(String& FileName);  //   .
    void StringToLog(String& message);
    bool ReallyLoadMob(String& FileName); //  .    .
    bool ReallySaveMob(String& FileName); //  .    .

    // .         .
    bool ReadScriptBlock(byte *Buff, DWORD& Size);  // .
    bool ReadObjectsBlock(byte *Buff, DWORD& Size);  // .

    // .               .
    bool GenScriptBlock(DynamicArray <byte>& Buff);  // .
    bool GenObjectsBlock(DynamicArray <byte>& Buff);  // .

    // .
    void KillUnit (const int ObjNumber);  // .
    void KillObject (const int ObjNumber);  // .
    void KillLever (const int ObjNumber);  // .
    void KillTrap (const int ObjNumber);  // .
    void KillFlame (const int ObjNumber);  // .
    void KillParticle1 (const int ObjNumber);  //  1- .
    void KillParticle2 (const int ObjNumber);  //  2- .
    void KillParticle3 (const int ObjNumber);  //  3- .

    //    .
    bool CheckIDExists (const DWORD ID, bool DeleteIfExists);  //     ID   .     -  .
    DWORD FindCleanID(const bool IdForNPC);    //  ID.

  private:
    void UnloadMob(); //            .
    void ClearObjectsBlockInfo();  //       ...
    String DeCodeScript(byte *ScriptTextBytes,DWORD Size,DWORD Key); //    .       .
    byte* EnCodeScript(String ScriptTextStr,DWORD Size,DWORD Key); //    .        .
    DWORD GetNextBlockSgn(DWORD CurrSgn);   //      .

};

class TMobLoaderThread : public TThread
{
  private:
    String message;
  protected:
    void __fastcall Execute();
  public:
    String FileName;
    TMob *PMobObj;
    bool Finished, result;
    __fastcall TMobLoaderThread(bool CreateSuspended);
    void SynchrLogMess(String& message);
    void __fastcall SynchrMessFunct(void);
    void __fastcall SynchrFinished(void);
};

class TMobSaverThread : public TThread
{
  private:
    String message;
  protected:
    void __fastcall Execute();
  public:
    String FileName;
    TMob *PMobObj;
    bool Finished, result;
    __fastcall TMobSaverThread(bool CreateSuspended);
    void SynchrLogMess(String& message);
    void __fastcall SynchrMessFunct(void);
    void __fastcall SynchrFinished(void);
};

//byte* DeCodeScript(byte *
//  ...
DWORD GetNMobBlocksSize(byte *Buff, DWORD StartPosition, DWORD NBlocks);   //    N    ,     -  1  .

#endif
